@@ -1,7 +1,7 @@
 /*
     Qalculate (GTK+ UI)
 
-    Copyright (C) 2003-2007, 2008, 2016  Hanna Knutsson (hanna_k@fmgirl.com)
+    Copyright (C) 2003-2007, 2008, 2016  Hanna Knutsson (hanna.knutsson@protonmail.com)
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -34,6 +34,9 @@ struct mode_struct {
 	int precision;
 	string name;
 	bool rpn_mode;
+	bool interval;
+	bool adaptive_interval_display;
+	bool variable_units_enabled;
 };
 
 enum {
@@ -47,17 +50,40 @@ enum {
 	QALCULATE_HISTORY_WARNING,
 	QALCULATE_HISTORY_ERROR,
 	QALCULATE_HISTORY_OLD,
-	QALCULATE_HISTORY_REGISTER_MOVED
+	QALCULATE_HISTORY_REGISTER_MOVED,
+	QALCULATE_HISTORY_RPN_OPERATION
 };
 
-bool can_display_unicode_string_function(const char *str, void *w);
-void set_unicode_buttons();
+DECLARE_BUILTIN_FUNCTION(AnswerFunction)
+DECLARE_BUILTIN_FUNCTION(ExpressionFunction)
 
-cairo_surface_t *draw_structure(MathStructure &m, PrintOptions po = default_print_options, InternalPrintStruct ips = top_ips, gint *point_central = NULL, int scaledown = 0);
+class ViewThread : public Thread {
+protected:
+	virtual void run();
+};
+
+class CommandThread : public Thread {
+protected:
+	virtual void run();
+};
+
+class FetchExchangeRatesThread : public Thread {
+protected:
+	virtual void run();
+};
+
+
+bool can_display_unicode_string_function(const char *str, void *w);
+bool can_display_unicode_string_function_exact(const char *str, void *w);
+void set_unicode_buttons();
+void set_operator_symbols();
+
+cairo_surface_t *draw_structure(MathStructure &m, PrintOptions po = default_print_options, InternalPrintStruct ips = top_ips, gint *point_central = NULL, int scaledown = 0, GdkRGBA *color = NULL);
 
 void update_status_text();
 
 void set_result_size_request();
+void set_expression_size_request();
 
 void create_umenu(void);
 void create_umenu2(void);
@@ -77,7 +103,7 @@ void generate_variables_tree_struct();
 void generate_units_tree_struct();
 
 gboolean on_display_errors_timeout(gpointer data);
-gboolean on_check_expression_position_timeout(gpointer data);
+gboolean on_check_version_idle(gpointer data);
 
 void update_functions_tree();
 void update_variables_tree();
@@ -163,6 +189,12 @@ void insert_variable(GtkMenuItem *w, gpointer user_data);
 void insert_prefix(GtkMenuItem *w, gpointer user_data);
 void insert_unit(GtkMenuItem *w, gpointer user_data);
 
+void insert_button_function(GtkMenuItem *w, gpointer user_data);
+void insert_button_function_norpn(GtkMenuItem *w, gpointer user_data);
+void insert_button_variable(GtkWidget *w, gpointer user_data);
+void insert_button_unit(GtkMenuItem *w, gpointer user_data);
+void insert_button_currency(GtkMenuItem *w, gpointer user_data);
+
 void new_function(GtkMenuItem *w, gpointer user_data);
 void new_unknown(GtkMenuItem *w, gpointer user_data);
 void new_variable(GtkMenuItem *w, gpointer user_data);
@@ -171,6 +203,8 @@ void new_vector(GtkMenuItem *w, gpointer user_data);
 void new_unit(GtkMenuItem *w, gpointer user_data);
 void add_as_variable();
 
+void fetch_exchange_rates(int timeout);
+
 void edit_argument(Argument *arg = NULL);
 
 void edit_unknown(const char *category = "", Variable *v = NULL, GtkWidget *win = NULL);
@@ -185,20 +219,22 @@ void edit_dataset(DataSet *ds = NULL, GtkWidget *win = NULL);
 
 void edit_names(ExpressionItem *item = NULL, const gchar *namestr = NULL, GtkWidget *win = NULL, bool is_dp = false, DataProperty *dp = NULL);
 
+gchar *font_name_to_css(const char *font_name);
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-gboolean on_completion_match_selected(GtkEntryCompletion *entrycompletion, GtkTreeModel *model, GtkTreeIter *iter, gpointer user_data);
-gboolean completion_match_func(GtkEntryCompletion *entrycompletion, const gchar *key, GtkTreeIter *iter, gpointer user_data);
+void on_completion_match_selected(GtkTreeView*, GtkTreePath *path, GtkTreeViewColumn*, gpointer);
 
 void *view_proc(void*);
 void *command_proc(void*);
 void on_message_bar_response(GtkInfoBar *w, gint response_id, gpointer);
-void on_history_populate_popup(GtkTextView *w, GtkMenu *menu, gpointer user_data);
-void on_expression_populate_popup(GtkEntry *w, GtkMenu *menu, gpointer user_data);
+void on_expressiontext_populate_popup(GtkTextView *w, GtkMenu *menu, gpointer user_data);
 void on_combobox_base_changed(GtkComboBox *w, gpointer user_data);
 void on_combobox_numerical_display_changed(GtkComboBox *w, gpointer user_data);
+void on_combobox_fraction_mode_changed(GtkComboBox *w, gpointer user_data);
+void on_combobox_approximation_changed(GtkComboBox *w, gpointer user_data);
 void on_expander_keypad_expanded(GObject *o, GParamSpec *param_spec, gpointer user_data);
 void on_expander_history_expanded(GObject *o, GParamSpec *param_spec, gpointer user_data);
 void on_expander_stack_expanded(GObject *o, GParamSpec *param_spec, gpointer user_data);
@@ -209,8 +245,8 @@ void on_menu_item_meta_mode_delete_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_quit_activate(GtkMenuItem *w, gpointer user_data);
 void on_colorbutton_status_error_color_color_set(GtkColorButton *w, gpointer user_data);
 void on_colorbutton_status_warning_color_color_set(GtkColorButton *w, gpointer user_data);
-void on_preferences_entry_wget_args_changed(GtkEditable *w, gpointer user_data);
 void on_preferences_checkbutton_lower_case_numbers_toggled(GtkToggleButton *w, gpointer user_data);
+void on_preferences_checkbutton_e_notation_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_checkbutton_lower_case_e_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_checkbutton_alternative_base_prefixes_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_checkbutton_spell_out_logical_operators_toggled(GtkToggleButton *w, gpointer user_data);
@@ -220,7 +256,7 @@ void on_preferences_checkbutton_enable_completion_toggled(GtkToggleButton *w, gp
 void on_preferences_checkbutton_fetch_exchange_rates_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_checkbutton_save_defs_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_checkbutton_save_mode_toggled(GtkToggleButton *w, gpointer user_data);
-void on_preferences_checkbutton_rpn_keypad_only_toggled(GtkToggleButton *w, gpointer);
+void on_preferences_checkbutton_rpn_keys_only_toggled(GtkToggleButton *w, gpointer);
 void on_preferences_checkbutton_dot_as_separator_toggled(GtkToggleButton *w, gpointer);
 void on_preferences_checkbutton_load_defs_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_checkbutton_custom_result_font_toggled(GtkToggleButton *w, gpointer user_data);
@@ -232,6 +268,9 @@ void on_preferences_radiobutton_asterisk_toggled(GtkToggleButton *w, gpointer us
 void on_preferences_radiobutton_slash_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_radiobutton_division_slash_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_radiobutton_division_toggled(GtkToggleButton *w, gpointer user_data);
+void on_preferences_radiobutton_digit_grouping_none_toggled(GtkToggleButton *w, gpointer user_data);
+void on_preferences_radiobutton_digit_grouping_standard_toggled(GtkToggleButton *w, gpointer user_data);
+void on_preferences_radiobutton_digit_grouping_locale_toggled(GtkToggleButton *w, gpointer user_data);
 void on_preferences_button_result_font_toggled(GtkButton *w, gpointer user_data);
 void on_preferences_button_expression_font_toggled(GtkButton *w, gpointer user_data);
 void on_preferences_button_status_font_toggled(GtkButton *w, gpointer user_data);
@@ -255,13 +294,12 @@ void on_button_ac_clicked(GtkButton *w, gpointer user_data);
 void on_button_hyp_toggled(GtkToggleButton *w, gpointer user_data);
 void on_button_inv_toggled(GtkToggleButton *w, gpointer user_data);
 void on_button_fraction_toggled(GtkToggleButton *w, gpointer user_data);
-void on_button_exact_toggled(GtkToggleButton *w, gpointer user_data);
 void on_button_tan_clicked(GtkButton *w, gpointer user_data);
 void on_button_sine_clicked(GtkButton *w, gpointer user_data);
 void on_button_cosine_clicked(GtkButton *w, gpointer user_data);
 void on_button_store_clicked(GtkButton *w, gpointer user_data);
 void on_button_mod_clicked(GtkButton *w, gpointer user_data);
-void on_button_factorial_clicked(GtkButton *w, gpointer user_data);
+void on_button_reciprocal_clicked(GtkButton *w, gpointer user_data);
 void on_togglebutton_expression_toggled(GtkToggleButton *togglebutton, gpointer user_data);
 void on_togglebutton_result_toggled(GtkToggleButton *togglebutton, gpointer user_data);
 void on_expression_changed(GtkEditable *w, gpointer user_data);
@@ -286,7 +324,7 @@ void on_button_divide_clicked(GtkButton *w, gpointer user_data);
 void on_button_ans_clicked(GtkButton *w, gpointer user_data);
 void on_button_exp_clicked(GtkButton *w, gpointer user_data);
 void on_button_xy_clicked(GtkButton *w, gpointer user_data);
-void on_button_square_clicked(GtkButton *w, gpointer user_data);
+void on_button_square_clicked();
 void on_button_sqrt_clicked(GtkButton *w, gpointer user_data);
 void on_button_log_clicked(GtkButton *w, gpointer user_data);
 void on_button_ln_clicked(GtkButton *w, gpointer user_data);
@@ -307,6 +345,7 @@ void on_menu_item_enable_functions_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_enable_units_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_enable_unknown_variables_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_calculate_variables_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_enable_variable_units_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_allow_complex_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_allow_infinite_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_new_unknown_activate(GtkMenuItem *w, gpointer user_data);
@@ -322,7 +361,6 @@ void on_menu_item_limit_implicit_multiplication_activate(GtkMenuItem *w, gpointe
 void on_menu_item_adaptive_parsing_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_ignore_whitespace_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_no_special_implicit_multiplication_activate(GtkMenuItem *w, gpointer user_data);
-void fetch_exchange_rates(int timeout);
 void on_menu_item_fetch_exchange_rates_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_save_defs_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_save_mode_activate(GtkMenuItem *w, gpointer user_data);
@@ -336,6 +374,7 @@ void output_base_updated_from_menu();
 void on_menu_item_binary_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_octal_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_decimal_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_duodecimal_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_hexadecimal_activate(GtkMenuItem *w, gpointer user_data);
 void on_number_base_spinbutton_base_value_changed(GtkSpinButton *w, gpointer user_data);
 void on_menu_item_custom_base_activate(GtkMenuItem *w, gpointer user_data);
@@ -346,6 +385,7 @@ void on_menu_item_set_base_activate(GtkMenuItem *w, gpointer user_data);
 void on_set_base_radiobutton_output_binary_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_output_octal_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_output_decimal_toggled(GtkToggleButton *w, gpointer user_data);
+void on_set_base_radiobutton_output_duodecimal_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_output_hexadecimal_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_output_other_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_output_sexagesimal_toggled(GtkToggleButton *w, gpointer user_data);
@@ -355,6 +395,7 @@ void on_set_base_spinbutton_output_other_value_changed(GtkSpinButton *w, gpointe
 void on_set_base_radiobutton_input_binary_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_input_octal_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_input_decimal_toggled(GtkToggleButton *w, gpointer user_data);
+void on_set_base_radiobutton_input_duodecimal_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_input_hexadecimal_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_input_other_toggled(GtkToggleButton *w, gpointer user_data);
 void on_set_base_radiobutton_input_roman_toggled(GtkToggleButton *w, gpointer user_data);
@@ -377,6 +418,7 @@ void on_menu_item_assumptions_nonpositive_activate(GtkMenuItem *w, gpointer user
 void on_menu_item_assumptions_unknown_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_assume_nonzero_denominators_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_abort_activate(GtkMenuItem *w, gpointer user_data);
+void on_popup_menu_item_clear_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_display_normal_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_display_engineering_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_display_scientific_activate(GtkMenuItem *w, gpointer user_data);
@@ -396,6 +438,7 @@ void on_popup_menu_item_sexagesimal_activate(GtkMenuItem *w, gpointer user_data)
 void on_popup_menu_item_time_format_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_octal_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_decimal_activate(GtkMenuItem *w, gpointer user_data);
+void on_popup_menu_item_duodecimal_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_hexadecimal_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_custom_base_activate(GtkMenuItem *w, gpointer user_data);
 void on_popup_menu_item_abbreviate_names_activate(GtkMenuItem *w, gpointer user_data);
@@ -421,16 +464,23 @@ void on_menu_item_fraction_decimal_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_fraction_decimal_exact_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_fraction_combined_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_fraction_fraction_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_interval_adaptive_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_interval_significant_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_interval_interval_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_interval_plusminus_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_interval_midpoint_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_all_prefixes_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_denominator_prefixes_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_place_units_separately_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_post_conversion_none_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_post_conversion_base_activate(GtkMenuItem *w, gpointer user_data);
-void on_menu_item_post_conversion_best_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_post_conversion_optimal_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_post_conversion_optimal_si_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_abbreviate_names_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_always_exact_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_try_exact_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_approximate_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_interval_arithmetic_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_save_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_save_image_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_copy_activate(GtkMenuItem *w, gpointer user_data);
@@ -443,7 +493,7 @@ void on_menu_item_set_unknowns_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_assume_nonzero_denominators_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_algebraic_mode_simplify_activate(GtkMenuItem *w, gpointer user_data);
 void on_menu_item_algebraic_mode_factorize_activate(GtkMenuItem *w, gpointer user_data);
-void on_menu_item_algebraic_mode_none_activate(GtkMenuItem *w, gpointer user_data);
+void on_menu_item_algebraic_mode_hybrid_activate(GtkMenuItem *w, gpointer user_data);
 gboolean on_main_window_focus_in_event(GtkWidget *w, GdkEventFocus *e, gpointer user_data);
 
 void on_button_registerup_clicked(GtkButton *button, gpointer user_data);
@@ -456,8 +506,11 @@ void on_stackview_item_edited(GtkCellRendererText *renderer, gchar *path, gchar
 void on_stackview_item_editing_started(GtkCellRenderer *renderer, GtkCellEditable *editable, gchar *path, gpointer user_data);
 void on_stackview_item_editing_canceled(GtkCellRenderer *renderer, gpointer user_data);
 
+void on_historyview_selection_changed(GtkTreeSelection *select, gpointer);
+
 void on_unit_edit_entry_name_changed(GtkEditable *editable, gpointer user_data);
 void on_unit_edit_combobox_class_changed(GtkComboBox *om, gpointer user_data);
+void on_unit_edit_checkbutton_mix_toggled(GtkToggleButton *w, gpointer);
 void on_unit_edit_combo_system_changed(GtkComboBox *om, gpointer user_data);
 void on_units_button_new_clicked(GtkButton *button, gpointer user_data);
 void on_units_button_edit_clicked(GtkButton *button, gpointer user_data);
@@ -503,6 +556,10 @@ void on_nbases_entry_binary_changed(GtkEditable *editable, gpointer user_data);
 void on_nbases_entry_octal_changed(GtkEditable *editable, gpointer user_data);
 void on_nbases_entry_hexadecimal_changed(GtkEditable *editable, gpointer user_data);
 
+void on_percentage_button_clear_clicked(GtkWidget*, gpointer);
+
+void on_menu_item_show_percentage_dialog_activate(GtkMenuItem *w, gpointer user_data);
+
 void on_button_functions_clicked(GtkButton *button, gpointer user_data);
 void on_button_variables_clicked(GtkButton *button, gpointer user_data);
 void on_button_units_clicked(GtkButton *button, gpointer user_data);
@@ -518,12 +575,14 @@ void on_decimals_dialog_spinbutton_min_value_changed(GtkSpinButton *w, gpointer
 void on_decimals_dialog_checkbutton_max_toggled(GtkToggleButton *w, gpointer user_data);
 void on_decimals_dialog_checkbutton_min_toggled(GtkToggleButton *w, gpointer user_data);
 
-gboolean on_expression_button_press_event(GtkWidget *w, GdkEventButton *event, gpointer user_data);
+gboolean on_expressiontext_button_press_event(GtkWidget *w, GdkEventButton *event, gpointer user_data);
 
 gboolean on_resultview_button_press_event(GtkWidget *w, GdkEventButton *event, gpointer user_data);
 gboolean on_resultview_popup_menu(GtkWidget *w, gpointer user_data);
 
-gboolean on_expression_key_press_event(GtkWidget *w, GdkEventKey *event, gpointer user_data);
+gboolean on_expressiontext_key_press_event(GtkWidget *w, GdkEventKey *event, gpointer user_data);
+
+void on_expressionbuffer_cursor_position_notify();
 
 gboolean on_resultview_draw(GtkWidget *w, cairo_t *cr, gpointer user_data);
 
