@@ -1,7 +1,7 @@
 /*
     Qalculate (GTK+ UI)
 
-    Copyright (C) 2003-2007, 2008, 2016  Hanna Knutsson (hanna_k@fmgirl.com)
+    Copyright (C) 2003-2007, 2008, 2016  Hanna Knutsson (hanna.knutsson@protonmail.com)
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,7 +16,9 @@
 
 #include <gtk/gtk.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
+#include <glib/gstdio.h>
 #include <unistd.h>
+#include <sys/stat.h>
 
 #include "support.h"
 #include "interface.h"
@@ -24,7 +26,6 @@
 #include "main.h"
 
 MathStructure *mstruct, *matrix_mstruct, *parsed_mstruct, *parsed_tostruct, *displayed_mstruct;
-bool prev_result_approx;
 string *parsed_to_str;
 KnownVariable *vans[5];
 GtkWidget *functions_window;
@@ -38,7 +39,7 @@ GtkWidget *units_window;
 string selected_unit_category;
 string selected_unit_selector_category;
 Unit *selected_unit, *selected_to_unit;
-bool load_global_defs, fetch_exchange_rates_at_startup, first_time, showing_first_time_message;
+bool load_global_defs, fetch_exchange_rates_at_startup, first_time, showing_first_time_message, allow_multiple_instances;
 cairo_surface_t *surface_result;
 GdkPixbuf *pixbuf_result;
 extern bool b_busy, b_busy_command, b_busy_result, b_busy_expression;
@@ -46,22 +47,25 @@ extern vector<string> recent_functions_pre;
 extern vector<string> recent_variables_pre;
 extern vector<string> recent_units_pre;
 extern GtkWidget *expression;
+extern GtkWidget *resultview;
 extern PrintOptions printops;
 
+MathFunction *f_answer;
+MathFunction *f_expression;
+
 GtkBuilder *main_builder, *argumentrules_builder, *csvimport_builder, *csvexport_builder, *setbase_builder, *datasetedit_builder, *datasets_builder, *decimals_builder;
 GtkBuilder *functionedit_builder, *functions_builder, *matrixedit_builder, *matrix_builder, *namesedit_builder, *nbases_builder, *plot_builder, *precision_builder;
 GtkBuilder *preferences_builder, *unit_builder, *unitedit_builder, *units_builder, *unknownedit_builder, *variableedit_builder, *variables_builder;
-GtkBuilder *periodictable_builder;
+GtkBuilder *periodictable_builder, *simplefunctionedit_builder, *percentage_builder;
 
-FILE *view_pipe_r, *view_pipe_w, *command_pipe_r, *command_pipe_w;
-pthread_t view_thread, command_thread;
-pthread_attr_t view_thread_attr, command_thread_attr;
-bool command_thread_started;
+Thread *view_thread, *command_thread;
 string calc_arg;
 
 bool do_timeout, check_expression_position;
 gint expression_position;
 
+QalculateDate last_version_check_date;
+
 static GOptionEntry options[] = {
 	{"new-instance", 'n', 0, G_OPTION_ARG_NONE, NULL, N_("Start a new instance of the application"), NULL},
 	{G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_STRING_ARRAY, NULL, N_("Expression to calculate"), N_("EXPRESSION")},
@@ -70,7 +74,18 @@ static GOptionEntry options[] = {
 
 void create_application(GtkApplication *app) {
 
-	gtk_window_set_default_icon_from_file(PACKAGE_DATA_DIR "/pixmaps/qalculate.png", NULL);
+#ifdef _WIN32
+	char exepath[MAX_PATH];
+	GetModuleFileName(NULL, exepath, MAX_PATH);
+	string datadir(exepath);
+	gchar *icon_path;
+	if(datadir.find("\\.libs") != string::npos) icon_path = g_build_filename(getPackageDataDir().c_str(), "data", "qalculate.png", NULL);
+	else icon_path = g_build_filename(getPackageDataDir().c_str(), "pixmaps", "qalculate.png", NULL);
+#else
+	gchar *icon_path = g_build_filename(getPackageDataDir().c_str(), "pixmaps", "qalculate.png", NULL);
+#endif
+	gtk_window_set_default_icon_from_file(icon_path, NULL);
+	g_free(icon_path);
 	
 	b_busy = false; 
 	b_busy_result = false;
@@ -80,9 +95,9 @@ void create_application(GtkApplication *app) {
 	main_builder = NULL; argumentrules_builder = NULL; 
 	csvimport_builder = NULL; datasetedit_builder = NULL; datasets_builder = NULL; decimals_builder = NULL; functionedit_builder = NULL; 
 	functions_builder = NULL; matrixedit_builder = NULL; matrix_builder = NULL; namesedit_builder = NULL; nbases_builder = NULL; plot_builder = NULL; 
-	precision_builder = NULL; preferences_builder = NULL; unit_builder = NULL; 
+	precision_builder = NULL; preferences_builder = NULL; unit_builder = NULL; percentage_builder = NULL; 
 	unitedit_builder = NULL; units_builder = NULL; unknownedit_builder = NULL; variableedit_builder = NULL; 
-	variables_builder = NULL;	csvexport_builder = NULL; setbase_builder = NULL; periodictable_builder = NULL;
+	variables_builder = NULL; csvexport_builder = NULL; setbase_builder = NULL; periodictable_builder = NULL, simplefunctionedit_builder = NULL;
 
 	//create the almighty Calculator object
 	new Calculator();
@@ -98,7 +113,6 @@ void create_application(GtkApplication *app) {
 	parsed_tostruct = new MathStructure();
 	parsed_tostruct->setUndefined();
 	matrix_mstruct = new MathStructure();
-	prev_result_approx = false;
 	parsed_to_str = new string;
 
 	bool canplot = CALCULATOR->canPlot();
@@ -108,46 +122,45 @@ void create_application(GtkApplication *app) {
 	create_main_window();
 
 	g_application_set_default(G_APPLICATION(app));
-	gtk_window_set_application(GTK_WINDOW(gtk_builder_get_object(main_builder, "main_window")), app);
+	gtk_window_set_application(GTK_WINDOW(gtk_builder_get_object(main_builder, "main_window")), app);	
 
 	while(gtk_events_pending()) gtk_main_iteration();
-	
+
 	showing_first_time_message = first_time;
 
 	if(!calc_arg.empty()) {
-		gtk_entry_set_text(GTK_ENTRY(gtk_builder_get_object(main_builder, "expression")), calc_arg.c_str());
+		gtk_text_buffer_set_text(GTK_TEXT_BUFFER(gtk_builder_get_object(main_builder, "expressionbuffer")), calc_arg.c_str(), -1);
 	} else if(first_time) {		
 		PangoLayout *layout = gtk_widget_create_pango_layout(GTK_WIDGET(gtk_builder_get_object(main_builder, "resultview")), NULL);
 		GdkRGBA rgba;
+#if GDK_MAJOR_VERSION > 3 || GDK_MINOR_VERSION >= 22
+		GdkDrawingContext *gdc = gdk_window_begin_draw_frame(gtk_widget_get_window(GTK_WIDGET(gtk_builder_get_object(main_builder, "resultview"))), cairo_region_create());
+		cairo_t *cr = gdk_drawing_context_get_cairo_context(gdc);
+#else
 		cairo_t *cr = gdk_cairo_create(gtk_widget_get_window(GTK_WIDGET(gtk_builder_get_object(main_builder, "resultview"))));
+#endif
 		pango_layout_set_markup(layout, _("<span font=\"10\">Type a mathematical expression above, e.g. \"5 + 2 / 3\",\nand press the enter key.</span>"), -1);
 		gtk_style_context_get_color(gtk_widget_get_style_context(GTK_WIDGET(gtk_builder_get_object(main_builder, "resultview"))), gtk_widget_get_state_flags(GTK_WIDGET(gtk_builder_get_object(main_builder, "resultview"))), &rgba);
 		cairo_move_to(cr, 6, 0);
 		gdk_cairo_set_source_rgba(cr, &rgba);
 		pango_cairo_show_layout(cr, layout);
 		g_object_unref(layout);
+#if GDK_MAJOR_VERSION > 3 || GDK_MINOR_VERSION >= 22
+		gdk_window_end_draw_frame(gtk_widget_get_window(GTK_WIDGET(gtk_builder_get_object(main_builder, "resultview"))), gdc);
+#else
 		cairo_destroy(cr);
+#endif
 	}
 
-	while(gtk_events_pending()) gtk_main_iteration();
+	while(gtk_events_pending()) gtk_main_iteration();	
 
 	//exchange rates
 	
 	if(fetch_exchange_rates_at_startup && canfetch) {
 		fetch_exchange_rates(5);
 		while(gtk_events_pending()) gtk_main_iteration();
-		CALCULATOR->loadExchangeRates();
-	} else if(!CALCULATOR->loadExchangeRates() && first_time && canfetch) {
-		GtkWidget *edialog = gtk_message_dialog_new(GTK_WINDOW(gtk_builder_get_object(main_builder, "main_window")), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO, _("You need to download exchange rates to be able to convert between different currencies. You can later get current exchange rates by selecting “Update Exchange Rates” under the File menu.\n\nDo you want to fetch exchange rates now from the Internet?"));
-		int question_answer = gtk_dialog_run(GTK_DIALOG(edialog));
-		gtk_widget_destroy(edialog);
-		while(gtk_events_pending()) gtk_main_iteration();
-		if(question_answer == GTK_RESPONSE_YES) {
-			fetch_exchange_rates(5);
-			while(gtk_events_pending()) gtk_main_iteration();
-			CALCULATOR->loadExchangeRates();
-		}
 	}
+	CALCULATOR->loadExchangeRates();
 
 	string ans_str = _("ans");
 	vans[0] = (KnownVariable*) CALCULATOR->addVariable(new KnownVariable(CALCULATOR->temporaryCategory(), ans_str, m_undefined, _("Last Answer"), false));
@@ -162,6 +175,9 @@ void create_application(GtkApplication *app) {
 	if(load_global_defs && !CALCULATOR->loadGlobalDefinitions()) {
 		g_print(_("Failed to load global definitions!\n"));
 	}
+	
+	f_answer = CALCULATOR->addFunction(new AnswerFunction());
+	f_expression = CALCULATOR->addFunction(new ExpressionFunction());
 
 	//load local definitions
 	CALCULATOR->loadLocalDefinitions();
@@ -184,24 +200,27 @@ void create_application(GtkApplication *app) {
 
 	//check for calculation errros regularly
 	do_timeout = true;
-	g_timeout_add(100, on_display_errors_timeout, NULL);
+	g_timeout_add_seconds(1, on_display_errors_timeout, NULL);
 	
 	check_expression_position = true;
 	expression_position = 1;
-	g_timeout_add(50, on_check_expression_position_timeout, NULL);
+	//g_timeout_add(50, on_check_expression_position_timeout, NULL);
 	
 	gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object (main_builder, "menu_item_plot_functions")), canplot);
 	gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object (main_builder, "menu_item_fetch_exchange_rates")), canfetch);
 
+	//create button menus after definitions have been loaded
+	create_button_menus();
+
 	//create dynamic menus
 	generate_units_tree_struct();
 
 	generate_functions_tree_struct();
 	generate_variables_tree_struct();
 	create_fmenu();
-	create_vmenu();	
+	create_vmenu();
 	create_umenu();
-	//create_pmenu();	
+	//create_pmenu();
 	create_umenu2();
 	create_pmenu2();
 
@@ -221,19 +240,9 @@ void create_application(GtkApplication *app) {
 
 	update_completion();
 	
-	int pipe_wr[] = {0, 0};
-	pipe(pipe_wr);
-	view_pipe_r = fdopen(pipe_wr[0], "r");
-	view_pipe_w = fdopen(pipe_wr[1], "w");
-	pthread_attr_init(&view_thread_attr);
-	pthread_create(&view_thread, &view_thread_attr, view_proc, view_pipe_r);
-	
-	int pipe_wr2[] = {0, 0};
-	pipe(pipe_wr2);
-	command_pipe_r = fdopen(pipe_wr2[0], "r");
-	command_pipe_w = fdopen(pipe_wr2[1], "w");
-	pthread_attr_init(&command_thread_attr);
-	command_thread_started = false;
+	view_thread = new ViewThread;
+	view_thread->start();
+	command_thread = new CommandThread;
 	
 	if(!calc_arg.empty()) {
 		execute_expression();
@@ -243,10 +252,29 @@ void create_application(GtkApplication *app) {
 		setResult(NULL, false, false, false);
 		printops.base = base;
 	}
+	gtk_widget_queue_draw(resultview);
 
-	gchar *gstr = g_build_filename(g_get_home_dir(), ".qalculate", "accelmap", NULL);
+	gchar *gstr = g_build_filename(getLocalDir().c_str(), "accelmap", NULL);
+#ifndef _WIN32
+	if(!fileExists(gstr)) {
+		g_free(gstr);
+		gstr = g_build_filename(getOldLocalDir().c_str(), "accelmap", NULL);
+		gtk_accel_map_load(gstr);
+		g_remove(gstr);
+		g_rmdir(getOldLocalDir().c_str());
+		g_free(gstr);
+		return;
+	}
+#endif
 	gtk_accel_map_load(gstr);
 	g_free(gstr);
+	
+#ifdef _WIN32
+	QalculateDate next_version_check_date(last_version_check_date);
+	next_version_check_date.addDays(14);
+	if(!next_version_check_date.isFutureDate()) g_idle_add(on_check_version_idle, NULL);
+#endif
+	
 }
 
 static void qalculate_activate(GtkApplication *app) {
@@ -259,8 +287,9 @@ static void qalculate_activate(GtkApplication *app) {
 		gtk_window_present(GTK_WINDOW(list->data));
 		return;
 	}
+
 	create_application(app);
-	
+
 }
 
 static gint qalculate_handle_local_options(GtkApplication *app, GVariantDict *options_dict) {
@@ -269,6 +298,53 @@ static gint qalculate_handle_local_options(GtkApplication *app, GVariantDict *op
 	if(b) {
 		g_application_set_flags(G_APPLICATION(app), G_APPLICATION_NON_UNIQUE);
 	}
+	allow_multiple_instances = false;
+	FILE *file = NULL;
+	gchar *gstr_oldfile = NULL;
+	gchar *gstr_file = g_build_filename(getLocalDir().c_str(), "qalculate-gtk.cfg", NULL);
+	file = fopen(gstr_file, "r");
+	if(!file) {
+#ifndef _WIN32
+		gstr_oldfile = g_build_filename(getOldLocalDir().c_str(), "qalculate-gtk.cfg", NULL);
+		file = fopen(gstr_oldfile, "r");
+		if(!file) {
+			g_free(gstr_oldfile);
+#endif
+			g_free(gstr_file);
+			return -1;
+#ifndef _WIN32
+		}
+#endif
+	}	
+	if(file) {
+		char line[100];
+		string stmp, svar;
+		size_t i;
+		while(true) {
+			if(fgets(line, 100, file) == NULL)
+				break;
+			stmp = line;
+			remove_blank_ends(stmp);
+			if((i = stmp.find_first_of("=")) != string::npos) {
+				svar = stmp.substr(0, i);
+				remove_blank_ends(svar);
+				if(svar == "allow_multiple_instances") {
+					string svalue = stmp.substr(i + 1, stmp.length() - (i + 1));
+					remove_blank_ends(svalue);
+					allow_multiple_instances = s2i(svalue);
+					break;
+				}
+			}
+		}
+		fclose(file);
+		if(gstr_oldfile) {
+			g_mkdir(getLocalDir().c_str(), S_IRWXU);
+			move_file(gstr_oldfile, gstr_file);
+			g_free(gstr_oldfile);
+		}
+	}
+	g_free(gstr_file);
+	if(allow_multiple_instances) g_application_set_flags(G_APPLICATION(app), G_APPLICATION_NON_UNIQUE);
 	return -1;
 }
 
@@ -291,7 +367,7 @@ static gint qalculate_command_line(GtkApplication *app, GApplicationCommandLine
 	if(main_builder) {
 		gtk_window_present(GTK_WINDOW(gtk_builder_get_object(main_builder, "main_window")));
 		if(!calc_arg.empty()) {
-			gtk_entry_set_text(GTK_ENTRY(gtk_builder_get_object(main_builder, "expression")), calc_arg.c_str());
+			gtk_text_buffer_set_text(GTK_TEXT_BUFFER(gtk_builder_get_object(main_builder, "expressionbuffer")), calc_arg.c_str(), -1);
 			execute_expression();
 		}
 	} else {
@@ -307,7 +383,7 @@ int main (int argc, char *argv[]) {
 	gint status;
 	
 #ifdef ENABLE_NLS
-	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
+	bindtextdomain (GETTEXT_PACKAGE, getPackageLocaleDir().c_str());
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 	textdomain (GETTEXT_PACKAGE);
 #endif
